/**
 * Copyright (C) 2013      Kai T. Ohlhus <k.ohlhus@gmail.com>
 * Copyright (C) 2012-2013 Wei Jin       <3162073@gmail.com>
 *
 * This file is part of Octave.
 *
 * Octave is free software; you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 3 of the License, or (at your option) any later
 * version.
 *
 * Octave is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * Octave; see the file COPYING.  If not, see <http://www.gnu.org/licenses/>.
 */

/**
 * This is the octave interface to itsol. For more information see
 * http://www-users.cs.umn.edu/~saad/software/ITSOL/
 */

//FIXME: remove when in octave-core
#include <octave/oct.h>
#define HAVE_ITSOL

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "defun-dld.h"
#include "error.h"
#include "gripes.h"
#include "oct-obj.h"
#include "utils.h"
#include "oct-map.h"

#include "MatrixType.h"
#include "SparseCmplxLU.h"
#include "SparsedbleLU.h"
#include "ov-re-sparse.h"
#include "ov-cx-sparse.h"
#include "Sparse.h"

// ITSOL libraries
extern "C"
{
#include <itsol/globheads.h>
#include <itsol/protos.h>
#include <itsol/defs.h>
#include <itsol/ios.h>
}

// Auxiliary ITSOL libraries
#include "itsol_util.h"

// System libraries
#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <cmath>
#include <ctime>

/*
  [L U] = iluk(A)
  [L U] = iluk(A,fill_lev)

  default value:
  1               1. fill_lev = fill level

  INPUT PARAMETERS
  ----------------------------------
  fill_lev    : Level of fill for ILUK preconditioner
*/
DEFUN_DLD (iluk, args, nargout, "  [L U] = iluk(A)\n\
  [L U] = iluk(A,fill_lev)\n\n\
  default value:\n\
  1               1. fill_lev = fill level\n\n\
  INPUT PARAMETERS\n\
  ----------------------------------\n\
  fill_lev    : Level of fill for ILUK preconditioner\n\n")
{
  octave_value_list retval;
#ifdef HAVE_ITSOL
  int nargin = args.length ();
  if (nargin == 0)
    print_usage ();
  else if (nargin > 2)
    error ("iluk: incorrect number of arguments");
  else if (nargin == 1)
    {
      if (args (0).type_name () == "sparse matrix")
	{
	  SparseMatrix sm = args (0).sparse_matrix_value ();
	  int *cidx = sm.cidx ();
	  int *ridx = sm.ridx ();
	  double *data = sm.data ();
	  int nnz = sm.nnz ();

	  csptr csmat = (csptr) malloc (sizeof (SparMat));
	  iluptr lu = (iluptr) malloc (sizeof (ILUSpar));
	  CSC_cs (sm.cols (), nnz, data, ridx, cidx, csmat);
	  int lfil = 1;
      /*-------------------- call ILUK preconditioner set-up  */
	  int ierr = ilukC (lfil, csmat, lu, stderr);
      /*----------------------------------------------------- */
	  if (ierr == -2)
	    {
	      cleanILU (lu);
	      error ("zero diagonal element found...\n");
	    }
	  else if (ierr != 0)
	    {
	      error ("*** iluk error, ierr != 0 ***\n");
	      exit (-1);
	    }
	  SparseMatrix U (lu->U->n, lu->U->n, nnz_cs (lu->U) + lu->U->n);
	  SparseMatrix L (lu->L->n, lu->L->n, nnz_cs (lu->L) + lu->L->n);
	  cs_CSC_U (U.data (), U.ridx (), U.cidx (), lu->U, lu->D);
	  cs_CSC_L (L.data (), L.ridx (), L.cidx (), lu->L);

	  cleanCS (csmat);
	  cleanILU (lu);
	  retval (0) = octave_value (L, MatrixType (MatrixType::Lower));
	  retval (1) = octave_value (U, MatrixType (MatrixType::Upper));
	  return retval;
	}
      else
	{
	  error ("First parameter should be sparse matrix");
	}
    }
  else if (nargin == 2)
    {
      if (args (0).type_name () == "sparse matrix")
	{
	  int lfil = args (1).int_value ();

	  if (!error_state)
	    {
	      SparseMatrix sm = args (0).sparse_matrix_value ();
	      int *cidx = sm.cidx ();
	      int *ridx = sm.ridx ();
	      double *data = sm.data ();

	      int nnz = sm.nnz ();

	      csptr csmat = (csptr) malloc (sizeof (SparMat));
	      iluptr lu = (iluptr) malloc (sizeof (ILUSpar));
	      CSC_cs (sm.cols (), nnz, data, ridx, cidx, csmat);
	/*-------------------- call ILUK preconditioner set-up  */
	      int ierr = ilukC (lfil, csmat, lu, stderr);
	/*----------------------------------------------------- */
	      if (ierr == -2)
		{
		  cleanILU (lu);
		  error ("zero diagonal element found...\n");
		}
	      else if (ierr != 0)
		{
		  error ("*** iluk error, ierr != 0 ***\n");
		  exit (-1);
		}
	      SparseMatrix U (lu->U->n, lu->U->n, nnz_cs (lu->U) + lu->U->n);
	      SparseMatrix L (lu->L->n, lu->L->n, nnz_cs (lu->L) + lu->L->n);
	      cs_CSC_U (U.data (), U.ridx (), U.cidx (), lu->U, lu->D);
	      cs_CSC_L (L.data (), L.ridx (), L.cidx (), lu->L);

	      cleanCS (csmat);
	      cleanILU (lu);
	      retval (0) = octave_value (L, MatrixType (MatrixType::Lower));
	      retval (1) = octave_value (U, MatrixType (MatrixType::Upper));
	      return retval;
	    }
	  else
	    {
	      error ("second parameter should be integer");
	    }
	}
      else
	{
	  error ("First parameter should be sparse matrix");
	}
    }
#else
  error ("iluk: ITSOL not installed");
#endif
  return retval;
}

