/**
 * Copyright (C) 2013      Kai T. Ohlhus <k.ohlhus@gmail.com>
 * Copyright (C) 2012-2013 Wei Jin       <3162073@gmail.com>
 *
 * This file is part of Octave.
 *
 * Octave is free software; you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 3 of the License, or (at your option) any later
 * version.
 *
 * Octave is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * Octave; see the file COPYING.  If not, see <http://www.gnu.org/licenses/>.
 */

/**
 * This is the octave interface to itsol. For more information see
 * http://www-users.cs.umn.edu/~saad/software/ITSOL/
 */

/**
 * This method transforms the Octave Compressed Sparse Column (CSC) (a.k.a.
 * compressed column storage (CCS)) via transformation to the coordinate format
 * (COO) to the ITSOL format (cs).
 */
void
CSC_cs (int n, int nnz, double *data, int *ridx, int *cidx, csptr bmat)
{
  // ja = cidx = col index
  int *ja = (int *) malloc (sizeof (int) * nnz);
  // ia = ridx = row index
  int *ia = (int *) malloc (sizeof (int) * nnz);

  int k = 0;
  for (int i = 0; i < n; i++)
    for (int j = 0; j < (cidx[i + 1] - cidx[i]); j++)
      {
	ja[k] = i;
	k++;
      }

  memcpy (ia, ridx, nnz * sizeof (int));	//FIXME: maybe avoid this copying like with data?
  COOcs (n, nnz, data, ja, ia, bmat);

  free (ia);
  free (ja);
}

void
csr_csc (int n, int nnz, double *a, int *ia, int *ja, double *ao, int *iao,
	 int *jao)
{
  //csr to csc
  //csr ia = n, ja = nnz, a = nnz;
  //csc iao = nnz, jao = n, ao = nnz;

  for (int i = 0; i < n + 1; i++)
    jao[i] = 0;
  for (int i = 0; i < nnz; i++)
    {
      jao[ja[i] + 1]++;
    }
  for (int i = 0; i < n; i++)
    {
      jao[i + 1] = jao[i + 1] + jao[i];
    }

  for (int i = 0; i < n; i++)
    {
      for (int j = ia[i]; j < ia[i + 1]; j++)
	{
	  ao[jao[ja[j]]] = a[j];
	  iao[jao[ja[j]]] = i;
	  jao[ja[j]]++;
	}
    }

  for (int i = n; i >= 0; i--)
    jao[i + 1] = jao[i];
  jao[0] = 0;
}

void
cs_CSC (double *ao, int *iao, int *jao, csptr bmat)
{
  //iao n+1, jao nnz
  //csc allocated array

  int nnz = nnz_cs (bmat);

  int *ia = (int *) malloc (sizeof (int) * (bmat->n + 1));
  int *ja = (int *) malloc (sizeof (int) * nnz);
  double *a = (double *) malloc (sizeof (double) * nnz);

  ia[0] = 0;
  for (int i = 0; i < bmat->n; i++)
    {
      ia[i + 1] = ia[i] + bmat->nzcount[i];
    }

  int k = 0;
  for (int i = 0; i < bmat->n; i++)
    {
      for (int j = 0; j < bmat->nzcount[i]; j++)
	{
	  ja[k] = bmat->ja[i][j];
	  a[k] = bmat->ma[i][j];
	  k++;
	}
    }
  csr_csc (bmat->n, nnz, a, ia, ja, ao, iao, jao);
  free (ia);
  free (ja);
  free (a);
}

void
cs_CSC_L (double *ao, int *iao, int *jao, csptr bmat)
{
  //iao n+1, jao nnz
  //csc allocated array

  int nnz = nnz_cs (bmat);

  int *ia = (int *) malloc (sizeof (int) * (bmat->n + 1));
  int *ja = (int *) malloc (sizeof (int) * (nnz + bmat->n));
  double *a = (double *) malloc (sizeof (double) * (nnz + bmat->n));

  ia[0] = 0;
  for (int i = 0; i < bmat->n; i++)
    {
      ia[i + 1] = ia[i] + bmat->nzcount[i] + 1;
    }

  int k = 0;
  for (int i = 0; i < bmat->n; i++)
    {
      for (int j = 0; j < bmat->nzcount[i]; j++)
	{
	  ja[k] = bmat->ja[i][j];
	  a[k] = bmat->ma[i][j];
	  k++;
	}
      ja[k] = i;
      a[k] = 1.0;
      k++;
    }
  csr_csc (bmat->n, nnz + bmat->n, a, ia, ja, ao, iao, jao);
  free (ia);
  free (ja);
  free (a);
}

void
cs_CSC_U (double *ao, int *iao, int *jao, csptr bmat, double *diag)
{
  //iao n+1, jao nnz
  //csc allocated array

  int nnz = nnz_cs (bmat);

  int *ia = (int *) malloc (sizeof (int) * (bmat->n + 1));
  int *ja = (int *) malloc (sizeof (int) * (nnz + bmat->n));
  double *a = (double *) malloc (sizeof (double) * (nnz + bmat->n));

  ia[0] = 0;
  for (int i = 0; i < bmat->n; i++)
    {
      ia[i + 1] = ia[i] + bmat->nzcount[i] + 1;
    }

  int k = 0;
  for (int i = 0; i < bmat->n; i++)
    {
      for (int j = 0; j < bmat->nzcount[i]; j++)
	{
	  ja[k] = bmat->ja[i][j];
	  a[k] = bmat->ma[i][j];
	  k++;
	}
      ja[k] = i;
      a[k] = 1.0 / diag[i];
      k++;
    }
  csr_csc (bmat->n, nnz + bmat->n, a, ia, ja, ao, iao, jao);
  free (ia);
  free (ja);
  free (a);
}

